/******************************************************************************
 *
 * Filename: double.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  4th March 2002
 *
 * Purpose:  Include file for double instructions.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

void d2f( Frame* fp )
{
	u4 valuew[2], resultw;
	double* d = (double*) valuew;
	float* f = (float*) &resultw;

	INSTRUCTION( "%p: d2f\n", fp->pc );

	POP_DOUBLE( fp, valuew );

	if( isDpInf(valuew) )  { PUSH_FLOAT( fp, fpInf ); return; }
	else if( isDnInf(valuew) )  { PUSH_FLOAT( fp, fnInf ); return; }
	else if( isDNaN(valuew) )  { PUSH_FLOAT( fp, fNaN ); return; }

	*f = (float) *d;
	PUSH_FLOAT( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: d2f: %f => %f\n", fp->pc, *d, *f ); )
}


void d2i( Frame* fp )
{
	u4 valuew[2], resultw;
	double* d = (double*) valuew;
	int32* i = (int32*) &resultw;
	int32 result;

	INSTRUCTION( "%p: d2l\n", fp->pc );

	POP_DOUBLE( fp, valuew );
	
	if( isDpInf(valuew) ) result = MAX_INTEGER;
	else if( isDnInf(valuew) ) result = MIN_INTEGER;
	else if( isDNaN(valuew) ) result = 0;
	else result = (int32) *d;

	*i = result;
	PUSH_INT( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: d2i: %f => %d\n", fp->pc, *d, (int) *i ); )
}


void d2l( Frame* fp )
{
	u4 valuew[2], resultw[2];
	double* d = (double*) valuew;
	int64* l = (int64*) resultw;
	int64 result;

	INSTRUCTION( "%p: d2l\n", fp->pc );

	POP_DOUBLE( fp, valuew );

	if( isDpInf(valuew) ) result = MAX_LONG;
	else if( isDnInf(valuew) ) result = MIN_LONG;
	else if( isDNaN(valuew) ) result = 0;
	else result = (int64) *d;

	*l = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: d2l: %f => 0x%08x%08x\n", fp->pc, *d, resultw[1], resultw[0] ); )
}


void dadd( Frame* fp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double* d  = (double*) resultw;
	double value1, value2, result;

	INSTRUCTION( "%p: dadd\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	result = value1 + value2;
	*d = result;

	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: dadd: %f + %f => %f\n", fp->pc, value1, value2, result ); )
}


int daload( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 index;
	ObjRef* array;
	u4 value[2];
	double* d = (double*) value;

	INSTRUCTION( "%p: daload\n", fp->pc );

	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "daload" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_DOUBLE) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad double[] object reference" );

	value[0] = array->data[index];
	value[1] = array->data[index+1];

	PUSH_DOUBLE( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: daload: 0x%p [ %d ] => %f\n", fp->pc, array->data, (int) index, *d ); )
	return 0;
}


int dastore( Stack* stp )
{
	Frame *fp = stp->fp;
	u4 value[2];
	int32 index;
	ObjRef* array;
	double* d = (double*) value;

	INSTRUCTION( "%p: dastore\n", fp->pc );

	POP_DOUBLE( fp, value );
	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "dastore" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_DOUBLE) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad double[] object reference" );

	array->data[index] = value[0];
	array->data[index+1] = value[1];

	ARITHMETIC_DEBUG( printf( "%p: dastore: 0x%p [ %d ] <= %f\n", fp->pc, array->data, (int) index, *d ); )
	return 0;
}


void dcmpg( Frame* fp )
{
	u4 valuew1[2], valuew2[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double value1, value2;
	int32 result = 0;

	INSTRUCTION( "%p: dcmpg\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	/* see section 7.5 */
	if( isDNaN(valuew1) || isDNaN(valuew2) )  { PUSH_INT( fp, 1 ); return; }

	if( value1 > value2 )
		result = 1;
	else if( value1 < value2 )
		result = -1;

	PUSH_INT( fp, result );

	ARITHMETIC_DEBUG( printf( "%p: dcmpg: %f == %f\n", fp->pc, value1, value2 ); )
}


void dcmpl( Frame* fp )
{
	u4 valuew1[2], valuew2[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double value1, value2;
	int32 result = 0;

	INSTRUCTION( "%p: dcmpl\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	/* see section 7.5 */
	if( isDNaN(valuew1) || isDNaN(valuew2) ) { PUSH_INT( fp, -1 ); return; }

	if( value1 > value2 )
		result = 1;
	else if( value1 < value2 )
		result = -1;

	PUSH_INT( fp, result );

	ARITHMETIC_DEBUG( printf( "%p: dcmpl: %f == %f\n", fp->pc, value1, value2 ); )
}


void dconst_n( Frame* fp, u1 n )
{
	u4 value[2];
	double* d = (double*) value;

	INSTRUCTION( "%p: dconst_%d\n", fp->pc, n );

	*d = (double) n;
	PUSH_DOUBLE( fp, value );
}


void ddiv( Frame* fp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double* d = (double*) resultw;
	double value1, value2, result;

	INSTRUCTION( "%p: ddiv\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	if( isDNaN( valuew2 ) )  { PUSH_DOUBLE( fp, dNaN ); return; }

	result = value1 / value2;
	*d = result;
	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: ddiv: %f / %f => %f\n", fp->pc, value1, value2, result ); )
}


void dload_n( Frame* fp, u2 n )
{
	INSTRUCTION( "%p: dload_%d\n", fp->pc, n );
	PUSH_LDOUBLE( fp, n );
}


void dmul( Frame* fp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double* d  = (double*) resultw;
	double value1, value2, result;

	INSTRUCTION( "%p: dmul\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	result = value1 * value2;
	*d = result;

	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: dmul: %f * %f => %f\n", fp->pc, value1, value2, result ); )
}


void dneg( Frame* fp )
{
	u4 valuew[2], resultw[2];
	double* lv = (double*) valuew;
	double* lr = (double*) resultw;
	double value, result;

	INSTRUCTION( "%p: dneg\n", fp->pc );

	POP_DOUBLE( fp, valuew );
	value = *lv;

	result = - value;
	*lr = result;

	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: dneg: %f => %f\n", fp->pc, value, result ); )
}


void dremainder( Frame *fp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double* r  = (double*) resultw;
	int64 v1, v2;
	double result;

	INSTRUCTION( "%p: frem\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	POP_DOUBLE( fp, valuew1 );

	if( isDNaN(valuew1) || isDNaN(valuew2) ) { PUSH_DOUBLE( fp, dNaN ); return; }
	if( isDpInf(valuew1) || isDnInf(valuew1) ) { PUSH_DOUBLE( fp, dNaN ); return; }
	if( isDpInf(valuew2) || isDnInf(valuew2) ) { PUSH_DOUBLE( fp, valuew1 ); return; }

	v1 = (int64) *d1;
	v2 = (int64) *d2;

	if( v2 == 0 ) { PUSH_DOUBLE( fp, dNaN ); return; }
	if( v1 == 0 ) { PUSH_DOUBLE( fp, valuew1 ); return; }

	result = (double) (v1 % v2);
	*r = result;

	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: drem: %f %% %f => %f\n", fp->pc, *d1, *d2, result ); )
}


int dreturn( Stack* stp )
{
	Frame* fp = stp->fp;
	u4 value[2];
	int stopOnPop = stp->fp->stopOnPop;

	INSTRUCTION( "%p: dreturn\n", fp->pc );

	POP_DOUBLE( fp, value );

	stp->fp = fp->next;
	freeStackFrame( fp, fp->mi );
	fp = stp->fp;

	if( fp )
	{
		fp->prev = NULL;
		PUSH_DOUBLE( fp, value );
	}

#ifdef _DEBUG
	if( fp && method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, fp->mi );
		printf( "Returned to method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( fp, 0 );

	return ( (!fp || stopOnPop) ? 1 : 0 );
}


void dstore_n( Frame* fp, u2 n )
{
	INSTRUCTION( "%p: dstore_%d\n", fp->pc, n );
	POP_LDOUBLE( fp, n );
}


void dsub( Frame* fp )
{
	u4 valuew1[2], valuew2[2], resultw[2];
	double* d1 = (double*) valuew1;
	double* d2 = (double*) valuew2;
	double* d  = (double*) resultw;
	double value1, value2, result;

	INSTRUCTION( "%p: dsub\n", fp->pc );

	POP_DOUBLE( fp, valuew2 );
	value2 = *d2;
	POP_DOUBLE( fp, valuew1 );
	value1 = *d1;

	result = value1 - value2;
	*d = result;

	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: dsub: %f - %f => %f\n", fp->pc, value1, value2, result ); )
}

