/******************************************************************************
 *
 * Filename: engine.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Execution engine interface and runtime datatypes.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#ifndef _engine_h

#define _engine_h


#include "classfile.h"


typedef enum _ExceptionKind
{
	/* Exception represented by enum value and message */
	VerifyError,			/* Invalid classfile */
	NoClassDefFoundError,	/* Couldn't load classfile */
	IllegalAccessError,		/* Can't access class/method/field */
	ClassFormatError,		/* Invalid classfile */
	NoSuchFieldError,		/* Field not found */
	NoSuchMethodError,		/* Method not found */
	InternalError,			/* VM error */
	UnsatisfiedLinkError,   /* Method not linked */

	NullPointerException,	/* null object reference */
	ArrayIndexOutOfBoundsException, /* Array access exception */
	ArithmeticException,    /* divide by zero, etc. */
	ClassCastException,		/* checkcast */
	ClassNotFoundException,	/* Couldn't load classfile */
	IOException,			/* input/output problems */
	NegativeArraySizeException, /* array size is invalid */

	/* Exception represented by ObjRef */
	JavaException			/* All Java exceptions */
} ExceptionKind;


/*
 * class java.lang.String
 * {
 *   private char value[];
 *   private int offset;
 *   private int count;
 * }
 */
typedef struct _javaLangString
{
	u4 value;
	u4 offset;
	u4 count;
} javaLangString;


/*
 * class java.lang.Thread
 * {
 *   private char		name[];
 *   private int        priority;
 *   private Thread		threadQ;
 *   private int 		PrivateInfo;
 *   private int		eetop;
 *   private boolean	single_step;
 *   private boolean	daemon;
 *   private boolean	stillborn;
 *   private Runnable	target;
 *   private ThreadGroup	group;
 *   private int		initial_stack_memory;
 * }
 */
typedef struct _javaLangThread
{
	u4 name;
	u4 priority;
	u4 threadQ;
	u4 PrivateInfo;
	u4 eetop;
	u4 single_step;
	u4 daemon;
	u4 stillborn;
	u4 target;
	u4 group;
	u4 initial_stack_memory;
} javaLangThread;


/*
 * class javaIoObjectStreamField
 * {
 *   String name;		// the name of the field
 *   char type;			// type first byte of the type signature
 *   int  offset;		// Offset into the object of the field
 *   String typeString;		// iff object, typename
 * }
 */
typedef struct _javaIoObjectStreamField
{
	u4 name;
	u4 type;
	u4 offset;
	u4 typeString;
} javaIoObjectStreamField;


typedef struct _ObjRef
{
	ClassFile* cf;			/* Object Type */
	u1 type;				/* Built-in Type */
	u4* data;				/* Data block */
	u4 length;				/* Array dimension */
} ObjRef;


typedef struct _ObjRefList
{
	ObjRef* obj;
	struct _ObjRefList* next;
} ObjRefList;


typedef struct _Exception
{
	ExceptionKind kind;
	char* message;
	ObjRef* ex;
} Exception;


typedef struct _Frame
{
	ClassFile* cf;
	method_info* mi;
	u1* pc;
	u1* min_pc;
	u1* max_pc;
	u4* local;
	u4* sp;
	u1* tlocal;
	u1* tsp;
	Exception* ex;
	struct _Frame* next;
	struct _Frame* prev;
	struct _Frame* init;
	int stopOnPop;

	/* data follows */
	u4* data;
} Frame;


typedef struct _Stack
{
	Exception* ex;
	Frame *fp;
	Frame *tail;
	int retcode;
} Stack;


/* pushes */
#define PUSH_OBJREF(fp,v)   *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) T_OBJREF; ((fp)->tsp)++;
#define PUSH_RETADDR(fp,v)  *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) T_RETADDR; ((fp)->tsp)++;
#define PUSH_INT(fp,v)      *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) T_INT; ((fp)->tsp)++;
#define PUSH_LONG(fp,v)     *((fp)->sp) = (u4) v[1]; ((fp)->sp)++; *((fp)->tsp) = (u1) T_LONG1; ((fp)->tsp)++; \
							*((fp)->sp) = (u4) v[0]; ((fp)->sp)++; *((fp)->tsp) = (u1) T_LONG; ((fp)->tsp)++;
#define PUSH_FLOAT(fp,v)    *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) T_FLOAT; ((fp)->tsp)++;
#define PUSH_DOUBLE(fp,v)   *((fp)->sp) = (u4) v[1]; ((fp)->sp)++; *((fp)->tsp) = (u1) T_DOUBLE1; ((fp)->tsp)++; \
							*((fp)->sp) = (u4) v[0]; ((fp)->sp)++; *((fp)->tsp) = (u1) T_DOUBLE; ((fp)->tsp)++;
#define PUSH_ANY_TYPE(fp,v,t)   *((fp)->sp) = (u4) v; ((fp)->sp)++; *((fp)->tsp) = (u1) t; ((fp)->tsp)++;
#define PUSH_LOBJREF(fp,n)  pushCheckLocal( fp, n, T_OBJREF );
#define PUSH_LINT(fp,n)     pushCheckLocal( fp, n, T_INT );
#define PUSH_LLONG(fp,n)    pushCheckLocal2( fp, n, T_LONG, T_LONG1 );
#define PUSH_LFLOAT(fp,n)   pushCheckLocal( fp, n, T_FLOAT );
#define PUSH_LDOUBLE(fp,n)  pushCheckLocal2( fp, n, T_DOUBLE, T_DOUBLE1 );
#define PUSH_LOCAL(fp,n)    *((fp)->sp) = (fp)->local[n]; ((fp)->sp)++; *((fp)->tsp) = (fp)->tlocal[n]; ((fp)->tsp)++;

/* pops */
#define POP_OBJREF(fp,v)   v = (ObjRef*) popCheck( fp, T_OBJREF );
#define POP_INT(fp,v)      v = (int32) popCheck( fp, T_INT );
#define POP_LONG(fp,v)     v[0] = popCheck( fp, T_LONG ); v[1] = popCheck1( fp, T_LONG1 );
#define POP_FLOAT(fp,v)    v = popCheck( fp, T_FLOAT );
#define POP_DOUBLE(fp,v)   v[0] = popCheck( fp, T_DOUBLE ); v[1] = popCheck1( fp, T_DOUBLE1 );
#define POP_ANY(fp,v)      v = popCheckAny( fp );
#define POP_ANY_TYPE(fp,v,t)   ((fp)->sp)--; v = *((fp)->sp); ((fp)->tsp)--; t = *((fp)->tsp);
#define POP_LOBJREF(fp,n)  popCheckLocal( fp, n, T_OBJREF );
#define POP_LINT(fp,n)     popCheckLocal( fp, n, T_INT );
#define POP_LLONG(fp,n)    popCheckLocal2( fp, n, T_LONG, T_LONG1 );
#define POP_LFLOAT(fp,n)   popCheckLocal( fp, n, T_FLOAT );
#define POP_LDOUBLE(fp,n)  popCheckLocal2( fp, n, T_DOUBLE, T_DOUBLE1 );
#define POP_LOCAL(fp,n)    ((fp)->sp)--; (fp)->local[n] = *((fp)->sp); ((fp)->tsp)--; (fp)->tlocal[n] = *((fp)->tsp);

/* local variable manipulation */
#define PUT_LOBJREF(fp,n,v)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) T_OBJREF;
#define GET_LRETADDR(fp,n,v)  v = getLocalRetAddr( fp, n );
#define PUT_LINT(fp,n,v)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) T_INT;
#define GET_LINT(fp,n,v)   v = getLocalInt( fp, n );
#define PUT_LANY_TYPE(fp,n,v,t)   (fp)->local[n] = (u4) v; (fp)->tlocal[n] = (u1) t;

/* misc */
#define DUP(fp)	  *((fp)->sp) = *((fp)->sp-1); ((fp)->sp)++; *((fp)->tsp) = *((fp)->tsp-1); ((fp)->tsp)++;

/* debug */
#ifdef _DEBUG
#define DEBUG(x) x
#define ARITHMETIC_DEBUG(x) if( arithmetic_debug ) { x }
#else
#define DEBUG(x) ;
#define ARITHMETIC_DEBUG(x) ;
#endif

#ifdef __cplusplus
extern "C" {
#endif


extern int subclassOf( const ClassFile* cf, const ClassFile* compare );

extern int newCompleteException( Stack *stp, int kind, const char* mess );

extern int forceExecuteJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc );
extern int forceExecuteJavaMethodWithArgs( ObjRef* objref, Stack* stp, const char* name, const char* desc, int argc, u4 argv[], u1 argvt[] );
extern int forceExecuteVirtualJavaMethod( ObjRef* objref, Stack* stp, const char* name, const char* desc, int stopOnPop );
extern int forceExecuteStaticJavaMethod( ClassFile* cf, Stack* stp, const char* name, const char* desc );
extern int executeCLInit( ClassFile* cf, Stack* stp );
extern int execute( ClassFile* cf, Stack* stp, int argc, char** argv );
extern int executeInstruction( Stack* stp );

extern u4 popCheck( Frame* fp, u1 type );
extern u4 popCheck1( Frame* fp, u1 type );
extern void pushCheckLocal( Frame* fp, u2 index, u1 type );
extern void pushCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 );
extern void popCheckLocal( Frame* fp, u2 index, u1 type );
extern void popCheckLocal2( Frame* fp, u2 index, u1 type0, u1 type1 );
extern ObjRef* getLocalObjRef( Frame* fp, u2 index );
extern int32 getLocalInt( Frame* fp, u2 index );

#ifdef __cplusplus
}
#endif


#endif
