/******************************************************************************
 *
 * Filename: float.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  4th March 2002
 *
 * Purpose:  Include file for float instructions.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

void f2d( Frame* fp )
{
	u4 valuew;
	u4 resultw[2] = { 0, 0 };
	float* f = (float*) &valuew;
	double* r = (double*) resultw;
	double result;

	INSTRUCTION( "%p: f2d\n", fp->pc );

	POP_FLOAT( fp, valuew );

	if( isFpInf(valuew) )  { PUSH_DOUBLE( fp, dpInf ); return; }
	else if( isFnInf(valuew) )  { PUSH_DOUBLE( fp, dnInf ); return; }
	else if( isFNaN(valuew) )  { PUSH_DOUBLE( fp, dNaN ); return; }

	result = (double) *f;
	*r = result;
	PUSH_DOUBLE( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: f2d: %f => %f\n", fp->pc, *f, result ); )
}


void f2i( Frame* fp )
{
	u4 valuew, resultw;
	float* f = (float*) &valuew;
	int32* r = (int32*) &resultw;
	int32 result;

	INSTRUCTION( "%p: f2i\n", fp->pc );

	POP_FLOAT( fp, valuew );

	if( isFpInf(valuew) ) result = MAX_INTEGER;
	else if( isFnInf(valuew) ) result = MIN_INTEGER;
	else if( isFNaN(valuew) ) result = 0;
	else result = (int32) *f;

	*r = result;
	PUSH_INT( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: f2i: %f => %d\n", fp->pc, *f, (int) result ); )
}


void f2l( Frame* fp )
{
	u4 valuew, resultw[2];
	float* f = (float*) &valuew;
	int64 result;
	int64* r = (int64*) resultw;

	INSTRUCTION( "%p: f2l\n", fp->pc );

	POP_FLOAT( fp, valuew );

	if( isFpInf(valuew) ) result = MAX_LONG;
	else if( isFnInf(valuew) ) result = MIN_LONG;
	else if( isFNaN(valuew) ) result = 0;
	else result = (int64) *f;

	*r = result;
	PUSH_LONG( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: f2l: %f => 0x%08x%08x\n", fp->pc, *f, resultw[1], resultw[0] ); )
}


void fadd( Frame* fp )
{
	u4 value;
	float* f = (float*) &value;
	float value1, value2, result;

	INSTRUCTION( "%p: fadd\n", fp->pc );

	POP_FLOAT( fp, value );
	value2 = *f;
	POP_FLOAT( fp, value );
	value1 = *f;

	result = value1 + value2;
	*f = result;
	PUSH_FLOAT( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: fadd: %f + %f => %f\n", fp->pc, (double) value1, (double) value2, (double) result ); )
}


int faload( Stack* stp )
{
	Frame *fp = stp->fp;
	int32 index;
	ObjRef* array;
	u4* data;

	INSTRUCTION( "%p: faload\n", fp->pc );

	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "faload" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_FLOAT) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad float[] object reference" );

	data = (u4*) array->data;

	PUSH_FLOAT( fp, data[index] );

	ARITHMETIC_DEBUG( printf( "%p: faload: 0x%p [ %d ] => %f\n", fp->pc, data, (int) index, (float) data[index] ); )
	return 0;
}


int fastore( Stack* stp )
{
	Frame *fp = stp->fp;
	u4 value;
	int32 index;
	ObjRef* array;
	u4* data;

	INSTRUCTION( "%p: iastore\n", fp->pc );

	POP_FLOAT( fp, value );
	POP_INT( fp, index );
	POP_OBJREF( fp, array );

	if( !array )
		return newCompleteException( stp, NullPointerException, "fastore" );

	if( (index < 0) || (index >= (int32) array->length) )
	{
		char buf[32];
		sprintf( buf, "%d", (int) index );
		return newCompleteException( stp, ArrayIndexOutOfBoundsException, buf );
	}

	if( (array->cf != NULL) || (array->type != T_FLOAT) || (array->data == NULL) )
		return newCompleteException( stp, InternalError, "Bad float[] object reference" );

	data = (u4*) array->data;

	data[index] = (u4) value;

	ARITHMETIC_DEBUG( printf( "%p: fastore: 0x%p [ %d ] <= %f\n", fp->pc, data, (int) index, (float) value ); )
	return 0;
}


void fcmpg( Frame* fp )
{
	u4 valuew1, valuew2;
	float* f1 = (float*) &valuew1;
	float* f2 = (float*) &valuew2;
	float value1, value2;
	int32 result = 0;

	INSTRUCTION( "%p: fcmpg\n", fp->pc );

	POP_FLOAT( fp, valuew2 );
	value2 = *f2;
	POP_FLOAT( fp, valuew1 );
	value1 = *f1;

	/* see section 7.5 */
	if( isFNaN(valuew1) || isFNaN(valuew2) ) { printf( "fcmpg: NaN encountered\n" ); PUSH_INT( fp, 1 ); return; }

	if( value1 > value2 )
		result = 1;
	else if( value1 < value2 )
		result = -1;

	PUSH_INT( fp, result );
}


void fcmpl( Frame* fp )
{
	u4 valuew1, valuew2;
	float* f1 = (float*) &valuew1;
	float* f2 = (float*) &valuew2;
	float value1, value2;
	int32 result = 0;

	INSTRUCTION( "%p: fcmpl\n", fp->pc );

	POP_FLOAT( fp, valuew2 );
	value2 = *f2;
	POP_FLOAT( fp, valuew1 );
	value1 = *f1;

	/* see section 7.5 */
	if( isFNaN(valuew1) || isFNaN(valuew2) ) { PUSH_INT( fp, -1 ); return; }

	if( value1 > value2 )
		result = 1;
	else if( value1 < value2 )
		result = -1;

	PUSH_INT( fp, result );
}


void fconst_n( Frame* fp, u1 n )
{
	u4 value;
	float* f = (float*) &value;

	INSTRUCTION( "%p: fconst_%d\n", fp->pc, n );

	*f = (float) n;

	PUSH_FLOAT( fp, value );
}


void fdiv( Frame* fp )
{
	u4 valuew1, valuew2, resultw;
	float* f1 = (float*) &valuew1;
	float* f2 = (float*) &valuew2;
	float* f = (float*) &resultw;
	float value1, value2, result;

	INSTRUCTION( "%p: fdiv\n", fp->pc );

	POP_FLOAT( fp, valuew2 );
	value2 = *f2;
	POP_FLOAT( fp, valuew1 );
	value1 = *f1;

	result = value1 / value2;
	*f = result;

	PUSH_FLOAT( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: fdiv: %f / %f => %f\n", fp->pc, value1, value2, result ); )
}


void fload_n( Frame* fp, u2 index )
{
	INSTRUCTION( "%p: fload_%d\n", fp->pc, index );
	PUSH_LFLOAT( fp, index );
}


void fmul( Frame* fp )
{
	u4 value;
	float* f = (float*) &value;
	float value1, value2, result;

	INSTRUCTION( "%p: fmul\n", fp->pc );

	POP_FLOAT( fp, value );
	value2 = *f;
	POP_FLOAT( fp, value );
	value1 = *f;

	result = value1 * value2;
	*f = result;
	PUSH_FLOAT( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: fmul: %f * %f => %f\n", fp->pc, value1, value2, result ); )
}


void fneg( Frame* fp )
{
	u4 value;
	float* f = (float*) &value;
	float value1, result;

	INSTRUCTION( "%p: fneg\n", fp->pc );

	POP_FLOAT( fp, value );
	value1 = *f;

	result = - value1;
	*f = result;
	PUSH_FLOAT( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: fneg: - %f => %f\n", fp->pc, value1, result ); )
}


void frem( Frame *fp )
{
	u4 valuew1, valuew2, resultw;
	float* f1 = (float*) &valuew1;
	float* f2 = (float*) &valuew2;
	float* f = (float*) &resultw;
	int32 v1, v2;
	float result;

	INSTRUCTION( "%p: frem\n", fp->pc );

	POP_FLOAT( fp, valuew2 );
	POP_FLOAT( fp, valuew1 );

	if( isFNaN(valuew1) || isFNaN(valuew2) ) { PUSH_FLOAT( fp, fNaN ); return; }
	if( isFpInf(valuew1) || isFnInf(valuew1) ) { PUSH_FLOAT( fp, fNaN ); return; }
	if( isFpInf(valuew2) || isFnInf(valuew2) ) { PUSH_FLOAT( fp, valuew1 ); return; }

	v1 = (int32) *f1;
	v2 = (int32) *f2;

	if( v2 == 0 ) { PUSH_FLOAT( fp, fNaN ); return; }
	if( v1 == 0 ) { PUSH_FLOAT( fp, valuew1 ); return; }

	result = (float) (v1 % v2);
	*f = result;

	PUSH_FLOAT( fp, resultw );

	ARITHMETIC_DEBUG( printf( "%p: frem: %f %% %f => %f\n", fp->pc, *f1, *f2, result ); )
}


int freturn( Stack* stp )
{
	Frame* fp = stp->fp;
	u4 value;
	int stopOnPop = stp->fp->stopOnPop;

	INSTRUCTION( "%p: freturn\n", fp->pc );

	POP_FLOAT( fp, value );

	stp->fp = fp->next;
	freeStackFrame( fp, fp->mi );
	fp = stp->fp;

	if( fp )
	{
		fp->prev = NULL;
		PUSH_FLOAT( fp, value );
	}

#ifdef _DEBUG
	if( fp && method_debug )
	{
		char buf[255];
		sprintMethodInfo( buf, fp->mi );
		printf( "Returned to method %s\n", buf );
	}
#endif /* _DEBUG */

	if( callstack )
		(*callstack)( fp, 0 );

	return ( (!fp || stopOnPop) ? 1 : 0 );
}


void fstore_n( Frame* fp, u2 index )
{
	INSTRUCTION( "%p: fstore_%d\n", fp->pc, index );
	POP_LFLOAT( fp, index );
}


void fsub( Frame* fp )
{
	u4 value;
	float* f = (float*) &value;
	float value1, value2, result;

	INSTRUCTION( "%p: fsub\n", fp->pc );

	POP_FLOAT( fp, value );
	value2 = *f;
	POP_FLOAT( fp, value );
	value1 = *f;

	result = value1 - value2;
	*f = result;
	PUSH_FLOAT( fp, value );

	ARITHMETIC_DEBUG( printf( "%p: fsub: %f - %f => %f\n", fp->pc, value1, value2, result ); )
}
