/******************************************************************************
 *
 * Filename: linker.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Static class linker.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "config.h"
#include "classfile.h"
#include "loader.h"
#include "verifier.h"
#include "disasm.h"
#include "native.h"
#include "memory.h"
#include "util.h"
#include "globals.h"

#include "linker.h"


extern void systemOutPrint( const char* buf, int len );

/*#define LOGFILE*/

#ifdef LOGFILE
FILE* log = NULL;
#endif


int compareUTF8( ClassFile* cf1, u2 index1, ClassFile* cf2, u2 index2 )
{
	if( (index1 > 0 ) && (index1 < cf1->constant_pool_count)
		&& (index2 > 0 ) && (index2 < cf2->constant_pool_count)
		&& cf1->constant_pool[index1] && (cf1->constant_pool[index1]->tag == CONSTANT_Utf8)
		&& cf2->constant_pool[index2] && (cf2->constant_pool[index2]->tag == CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp1 = (CONSTANT_Utf8_info*) cf1->constant_pool[index1];
		CONSTANT_Utf8_info* utfp2 = (CONSTANT_Utf8_info*) cf2->constant_pool[index2];

		/*printf( "Comparing \"" ); writeUTF8( utfp1->bytes, utfp1->length );
		printf( "\" to \"" ); writeUTF8( utfp2->bytes, utfp2->length ); printf( "\"\n" );*/ 

		if( utfp1->length != utfp2->length )
			return 0;

		return (strncmp( (const char*) utfp1->bytes, (const char*) utfp2->bytes, utfp1->length ) == 0);
	}

	return 0;
}


u1 countArguments( ClassFile* cf, u2 index )
{
	u1 args = 0;

	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[index];
		u2 n, array = 0;

		if( !utfp->length || (utfp->bytes[0] != '(') )
			return 0;

		for( n = 1; (n < utfp->length) && (utfp->bytes[n] != ')'); n++ )
		{
			switch( utfp->bytes[n] )
			{
			case 'B': args++; array = 0; break;
			case 'C': args++; array = 0; break;
			case 'D': if( array ) { args++; } else { args+=2; }; array = 0; break;
			case 'F': args++; array = 0; break;
			case 'I': args++; array = 0; break;
			case 'J': if( array ) { args++; } else { args+=2; }; array = 0; break;
			case 'L':
				args++; array = 0;
				for( n++; (n < utfp->length) && (utfp->bytes[n] != ';'); n++ );
				break;
			case 'S': args++; array = 0; break;
			case 'Z': args++; array = 0; break;
			case '[': array++; break;
			}
		}
	}

	return args;
}


u1 countField( ClassFile* cf, u2 index )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Utf8) )
	{
		CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[index];

		switch( utfp->bytes[0] )
		{
		case 'B': return 1;
		case 'C': return 1;
		case 'D': return 2;
		case 'F': return 1;
		case 'I': return 1;
		case 'J': return 2;
		case 'L': return 1;
		case 'S': return 1;
		case 'Z': return 1;
		case '[': return 1;
		}
	}

	return 0;
}


void resolveMethodAttributes( ClassFile* cf, method_info* mi )
{
	mi->cf = cf;	/* link method_info to ClassFile */
	mi->args = countArguments( cf, mi->descriptor_index );

	/* Process Native methods */
	if( mi->access_flags & ACC_NATIVE )
	{
		mi->nativestr = findNativeMethod( cf, mi->name_index, mi->descriptor_index );

		if( !mi->nativestr && 0 /* verbose ### */ )
		{
			char buf[255];
			sprintMethodInfo( buf, mi );
			printf( "LINKER: No native implementation of %s\n", buf );
		}

		return;
	}

	/* Process Java methods */
	if( mi->attributes_count )
	{
		u2 m;

		for( m = 0; m < mi->attributes_count; m++ )
		{
			if( mi->attributes[m] && checkSimpleName( cf, mi->attributes[m]->attribute_name_index, "Code" ) )
			{
				/* link Code_attribute and code itself */
				mi->ca = (Code_attribute*) mi->attributes[m];
				mi->nativestr = NULL;

				if( mi->ca->attributes_count )
				{
					u2 p;

					for( p = 0; p < mi->ca->attributes_count; p++ )
					{
						if( mi->ca->attributes[p] && checkSimpleName( cf, mi->ca->attributes[p]->attribute_name_index, "LineNumberTable" ) )
						{
							mi->lnta = (LineNumberTable_attribute*) mi->ca->attributes[p];
						}
						else if( mi->ca->attributes[p] && checkSimpleName( cf, mi->ca->attributes[p]->attribute_name_index, "LocalVariableTable" ) )
						{
							mi->lvta = (LocalVariableTable_attribute*) mi->ca->attributes[p];
						}
					}
				}
			}
		}

		/* Check we found some code */
		if( !mi->ca && !(mi->access_flags & ACC_ABSTRACT) )
		{
			char buf[255];
			sprintMethodInfo( buf, mi );
			printf( "LINKER: No Java code for %s\n", buf );
		}
	}
}


void linkMethodRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Methodref ) )
	{
		CONSTANT_Methodref_info* mp = (CONSTANT_Methodref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		method_info* mip;
		u2 m;

		if( !(CHECKCONSTANT( cf, mp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No Class constant for method %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[mp->class_index];

		if( !(CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for method %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, mp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, mp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->methods_count)
			&& !( compareUTF8( cp->cf, cp->cf->methods[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf, cp->cf->methods[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->methods_count )
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			printf( "LINKER: Failed to find method %s\n", buf );
			return;
		}

		mip = cp->cf->methods[m];
		
#ifdef LOGFILE
		{
			char buf[255];
			sprintMethodRef( buf, cf, index );
			fprintf( log, "Linker: Resolved method %s\n", buf );
		}
#endif
	
		mp->info = mip;		/* link Methodref to method_info */

		resolveMethodAttributes( cp->cf, cp->cf->methods[m] );
	}
}


void linkInterfaceMethodRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_InterfaceMethodref ) )
	{
		CONSTANT_InterfaceMethodref_info* mp = (CONSTANT_InterfaceMethodref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		method_info* mip;
		u2 m;

		if( !(CHECKCONSTANT( cf, mp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: No Class constant for method %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[mp->class_index];

		if( !(CHECKCONSTANT( cf, mp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for method %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[mp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, mp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, mp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->methods_count)
			&& !( compareUTF8( cp->cf, cp->cf->methods[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf, cp->cf->methods[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->methods_count )
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			printf( "LINKER: Failed to find method %s\n", buf );
			return;
		}

		mip = cp->cf->methods[m];
		
#ifdef LOGFILE
		{
			char buf[255];
			sprintInterfaceMethodRef( buf, cf, index );
			fprintf( log, "Linker: Resolved method %s\n", buf );
		}
#endif
	
		mp->info = mip;		/* link Methodref to method_info */

		resolveMethodAttributes( cp->cf, cp->cf->methods[m] );
	}
}


void linkFieldRef( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Fieldref ) )
	{
		CONSTANT_Fieldref_info* fp = (CONSTANT_Fieldref_info*) cf->constant_pool[index];
		CONSTANT_Class_info* cp;
		CONSTANT_NameAndType_info* natp;
		field_info* fip;
		u2 m;

		if( !(CHECKCONSTANT( cf, fp->class_index, CONSTANT_Class)) )
		{
			char buf[255];
			sprintFieldRef( buf, cf, index );
			printf( "LINKER: No Class constant for field %s\n", buf );
			return;
		}

		cp = (CONSTANT_Class_info*) cf->constant_pool[fp->class_index];

		if( !(CHECKCONSTANT( cf, fp->name_and_type_index, CONSTANT_NameAndType)) )
		{
			char buf[255];
			sprintFieldRef( buf, cf, index );
			printf( "LINKER: No NameAndType constant for field %s\n", buf );
			return;
		}

		natp = (CONSTANT_NameAndType_info*) cf->constant_pool[fp->name_and_type_index];
		
		if( !cp->cf )
		{
			linkClass( cf, fp->class_index, stp );

			if( stp->ex || (stp->fp && stp->fp->ex) || !cp->cf )
			{
				char buf[255];
				sprintClassName( buf, cf, fp->class_index );
				printf( "LINKER: Class \"%s\" cannot be loaded\n", buf );
				return;
			}
		}
		
		for( m = 0; (m < cp->cf->fields_count)
			&& !( compareUTF8( cp->cf, cp->cf->fields[m]->name_index, cf, natp->name_index )
				&& compareUTF8( cp->cf, cp->cf->fields[m]->descriptor_index, cf, natp->descriptor_index ) );
			m++ );

		if( m >= cp->cf->fields_count )
		{
			char buf[255];
			sprintFieldRef( buf, cf, index );
			printf( "LINKER: Failed to find field %s\n", buf );
			return;
		}

		fip = cp->cf->fields[m];
	
#ifdef LOGFILE
		{
			char buf[255];
			sprintFieldRef( buf, cf, index );
			fprintf( log, "Linker: Resolved field %s\n", buf );
		}
#endif
		
		fp->info = fip;		/* link Fieldref to method_info */
		fip->cf = cp->cf;	/* link field_info to ClassFile */
	}
}


void linkClass( ClassFile* cf, u2 index, Stack* stp )
{
	if( CHECKCONSTANT( cf, index, CONSTANT_Class ) )
	{
		CONSTANT_Class_info* cp = (CONSTANT_Class_info*) cf->constant_pool[index];

		if( CHECKCONSTANT( cf, cp->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[cp->name_index];
			char buf[255];
		
			strncpy( buf, (const char*) utfp->bytes, utfp->length );
			buf[utfp->length] = '\0';

			cp->cf = load( buf, stp );
		}
	}
}


/* resolve this class and all its super-classes */
void resolveSuperClasses( ClassFile* cf )
{
	u2 n, m;

#ifdef LOGFILE
	char classname[255];
	sprintClassName( classname, cf, cf->this_class );
	fprintf( log, "Resolving class \"%s\"...\n", classname );
#endif

	cf->resolved = 1;

	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n] && !cf->fields[n]->cf )
		{
			cf->fields[n]->cf = cf;
		}
	}

	for( n = 0; n < cf->methods_count; n++ )
	{
		if( cf->methods[n] && !cf->methods[n]->cf )
		{
			resolveMethodAttributes( cf, cf->methods[n] );
		}
	}

	/* resolve source file attribute */
	for( n = 0; n < cf->attributes_count; n++ )
	{
		if( cf->attributes[n] && checkSimpleName( cf, cf->attributes[n]->attribute_name_index, "SourceFile" ) )
		{
			cf->sfa = (SourceFile_attribute*) cf->attributes[n];
			break;
		}
	}

	/* Force base clase to be resolved first */
	if( cf->super )
	{
		if( !cf->super->resolved )
			resolveSuperClasses( cf->super );

		cf->object_size = cf->super->object_size;
	}
	else
		cf->object_size = 0;

	/* Calculate sizes of static and object data */
	for( n = 0; n < cf->fields_count; n++ )
	{
		if( cf->fields[n] )
		{
			if( cf->fields[n]->access_flags & ACC_STATIC )
			{
				cf->fields[n]->offset = cf->static_size;
				cf->static_size += countField( cf, cf->fields[n]->descriptor_index );
			}
			else
			{
				cf->fields[n]->offset = cf->object_size;
				cf->object_size += countField( cf, cf->fields[n]->descriptor_index );
			}
		}
	}

#ifdef LOGFILE
	fprintf( log, "Class \"%s\", object size = %d, static size = %d\n", classname, cf->object_size, cf->static_size );
#endif

	/* do virtual method resolution */
	if( !cf->super && !cf->interfaces_count )
	{
#ifdef LOGFILE
		fprintf( log, "Class \"%s\" has no super-classes.\n", classname );

		for( m = 0; m < cf->methods_count; m++ )
		{
			char mbuf[255];
			sprintMethodInfo( mbuf, cf->methods[m] );
			fprintf( log, "(%d) Added method \"%s\"\n", m, mbuf );
		}
#endif

		cf->virtuals = cf->methods;
		cf->virtuals_count = cf->methods_count;
		return;
	}

	cf->virtuals = javavm_malloc( cf->methods_count * sizeof( method_info* ) );
	cf->virtuals_count = 0;

	if( cf->super )
	{
#ifdef LOGFILE
		char sbuf[255];
		sprintClassName( sbuf, cf->super, cf->super->this_class );
		fprintf( log, "Class \"%s\" has super-class \"%s\".\n", classname, sbuf );
#endif

		if( cf->super->virtuals )
		{
			/* copy super-class' virtual method table */
			cf->virtuals = realloc( cf->virtuals, (cf->virtuals_count + cf->super->virtuals_count)* sizeof( method_info* ) );
			memcpy( &cf->virtuals[cf->virtuals_count], cf->super->virtuals, cf->super->virtuals_count * sizeof( method_info* ) );
#ifdef LOGFILE
			for( m = cf->virtuals_count; m < cf->virtuals_count + cf->super->virtuals_count; m++ )
			{
				char mbuf[255];
				sprintMethodInfo( mbuf, cf->virtuals[m] );
				fprintf( log, "(%d) Inherited method \"%s\"\n", m, mbuf );
			}
#endif
			cf->virtuals_count += cf->super->virtuals_count;
		}
		else
		{
			char cbuf[255], sbuf[255];
			sprintClassName( cbuf, cf, cf->this_class );
			sprintClassName( sbuf, cf, cf->super_class );
			printf( "LINKER: super-class \"%s\" of \"%s\" has not been resolved properly\n", cbuf, sbuf );
		}
	}

	if( cf->interfaces_count )
	{
		for( n = 0; n < cf->interfaces_count; n++ )
		{
			if( cf->interfaces[n] )
			{
#ifdef LOGFILE
				char sbuf[255];
				sprintClassName( sbuf, cf, cf->interfaces[n]->class_index );
				fprintf( log, "Class \"%s\" has super-interface \"%s\".\n", classname, sbuf );
#endif

				if( cf->interfaces[n]->cf && cf->interfaces[n]->cf->virtuals )
				{
					/* copy super-interface's virtual method table */
					cf->virtuals = realloc( cf->virtuals, (cf->virtuals_count + cf->interfaces[n]->cf->virtuals_count) * sizeof( method_info* ) );

					if( cf->virtuals_count > 0 )
					{
						u2 p;
						for( p = 0; p < cf->interfaces[n]->cf->virtuals_count; p++ )
						{
							for( m = 0; (m < cf->virtuals_count) && (cf->virtuals[m] != cf->interfaces[n]->cf->virtuals[p]); m++ );

							if( m >= cf->virtuals_count )
							{
#ifdef LOGFILE
								char mbuf[255];
								sprintMethodInfo( mbuf, cf->interfaces[n]->cf->virtuals[p] );
								fprintf( log, "(%d) Inherited method \"%s\"\n", cf->virtuals_count, mbuf );
#endif
								cf->virtuals[cf->virtuals_count] = cf->interfaces[n]->cf->virtuals[p];
								cf->virtuals_count++;
							}
						}
					}
					else
					{
						memcpy( &cf->virtuals[cf->virtuals_count], cf->interfaces[n]->cf->virtuals, cf->interfaces[n]->cf->virtuals_count * sizeof( method_info* ) );
#ifdef LOGFILE
						for( m = cf->virtuals_count; m < cf->virtuals_count + cf->interfaces[n]->cf->virtuals_count; m++ )
						{
							char mbuf[255];
							sprintMethodInfo( mbuf, cf->virtuals[m] );
							fprintf( log, "(%d) Inherited method \"%s\"\n", m, mbuf );
						}
#endif
						cf->virtuals_count += cf->interfaces[n]->cf->virtuals_count;
					}
				}
				else
				{
					char cbuf[255], ibuf[255];
					sprintClassName( cbuf, cf, cf->this_class );
					sprintClassName( ibuf, cf, cf->interfaces[n]->class_index );
					printf( "LINKER: super-interface \"%s\" of \"%s\" has not been resolved properly\n", cbuf, ibuf );
				}
			}
		}
	}

#ifdef LOGFILE
	fprintf( log, "Class \"%s\" has %d virtuals before overrides.\n", classname, cf->virtuals_count );
#endif

	/* process overrides */
	for( n = 0; n < cf->virtuals_count; n++ )
	{
		ClassFile *vcf = cf->virtuals[n]->cf;
		method_info *vmi = cf->virtuals[n];

		for( m = 0; (m < cf->methods_count)
			&& !( compareUTF8( cf, cf->methods[m]->name_index, vcf, vmi->name_index )
				&& compareUTF8( cf, cf->methods[m]->descriptor_index, vcf, vmi->descriptor_index ) );
			m++ );

		if( m < cf->methods_count )
		{
#ifdef LOGFILE
			char mbuf[255], vbuf[255];
			sprintMethodInfo( mbuf, cf->methods[m] );
			sprintMethodInfo( vbuf, cf->virtuals[n] );
			fprintf( log, "(%d) Resolved virtual method \"%s\" over \"%s\"\n", n, mbuf, vbuf );
#endif
			cf->virtuals[n] = cf->methods[m];
			cf->methods[m]->index = n;
		}
	}

	cf->virtuals = realloc( cf->virtuals, (cf->virtuals_count + cf->methods_count) * sizeof( method_info* ) );

	/* add this classes methods to end of virtual table */
	for( n = 0; n < cf->methods_count; n++ )
	{
		for( m = 0; (m < cf->virtuals_count) && (cf->virtuals[m] != cf->methods[n]); m++ );

		if( m >= cf->virtuals_count )
		{
#ifdef LOGFILE
			char mbuf[255];
			sprintMethodInfo( mbuf, cf->methods[n] );
			fprintf( log, "(%d) Added method \"%s\"\n", cf->virtuals_count, mbuf );
#endif
			cf->virtuals[cf->virtuals_count] = cf->methods[n];
			cf->methods[n]->index = cf->virtuals_count;
			cf->virtuals_count++;
		}
	}

#ifdef LOGFILE
	fprintf( log, "Class \"%s\" has %d virtuals after overrides.\n", classname, cf->virtuals_count );
#endif
}


ClassFile* loadClass( const char* classname, ClassFile* cf, u2 index, Stack* stp )
{
	char buf[255];

	if( CHECKCONSTANT( cf, index, CONSTANT_Class) )
	{
		CONSTANT_Class_info* cip = (CONSTANT_Class_info*) cf->constant_pool[index];

		if( CHECKCONSTANT( cf, cip->name_index, CONSTANT_Utf8 ) )
		{
			CONSTANT_Utf8_info* utfp = (CONSTANT_Utf8_info*) cf->constant_pool[cip->name_index];

			strncpy( buf, (const char*) utfp->bytes, utfp->length );
			buf[utfp->length] = '\0';

			if( buf[0] == '[' )
			{
				stp->retcode = newCompleteException( stp, NoClassDefFoundError, buf );
				return NULL;
			}

			if( !(cip->cf = load( buf, stp )) )
			{
				stp->retcode = newCompleteException( stp, NoClassDefFoundError, buf );
				return NULL;
			}

			return cip->cf;
		}
	}

	sprintf( buf, "Super-interface constant pool entry corrupt in class \"%s\"", classname );
	stp->retcode = newCompleteException( stp, InternalError, buf );
	return NULL;
}


ClassFile* load( const char* classname, Stack* stp )
{
	int retcode = 0;
	ClassFile* cf;
	char filename[255];
	char *buf = filename;

    cf = checkLoaded( classname );
    
    if( cf )
        return cf;

#ifdef USES_ABSOLUTE_FILENAMES
    strcpy( buf, ABSOLUTE_FILENAME_ROOT ); buf += strlen( ABSOLUTE_FILENAME_ROOT );
#endif /* USES_ABSOLUTE_FILENAMES */

	strcpy( buf, classname ); buf += strlen( classname );
	strcpy( buf, CLASS_FILE_EXTN ); buf += strlen( CLASS_FILE_EXTN );
	*buf = '\0';
	
#ifdef USES_BACKSLASHED_FILENAMES
    for( buf = filename; *buf; buf++ )
    {
        if( *buf == '/' ) *buf = '\\';
    }
#endif /* USES_BACKSLASHED_FILENAMES */
	
#ifdef USES_NONDOLLAR_FILENAMES
    for( buf = filename; *buf; buf++ )
    {
        if( *buf == '$' ) *buf = '_';
    }
#endif /* USES_NONDOLLAR_FILENAMES */

	if( !(cf = readClassFile( filename, classname, stp )) )
		return NULL; /* load failed - exception will be on stack */

	if( cf->interfaces_count )
	{
		u2 n;

		for( n = 0; n < cf->interfaces_count; n++ )
		{
			if( cf->interfaces[n] )
			{
				if( !(cf->interfaces[n]->cf = loadClass( classname, cf, cf->interfaces[n]->class_index, stp )) )
					return NULL;
			}
		}
	}

	if( cf->super_class )
	{
		ClassFile* scp;

		if( !(scp = loadClass( classname, cf, cf->super_class, stp )) )
			return NULL;

		cf->super = scp;

#ifdef LOGFILE
		log = fopen( "./javavm.log", "aw" );
#endif

		resolveSuperClasses( cf );

#ifdef LOGFILE
		fclose( log );
#endif
	}

	if( !verifyPass2( cf, stp ) )
		return NULL; /* verify puts exception on stack */

	if( !cf->initialised )
		retcode = initialise( cf, stp );

	if( retcode == 2 )
	{
		printf( "EXCEPTION: Exception in initialiser\n" );
	}

	/* ### should really be exception in initialiser */
	return (retcode == 2 ? NULL : cf);
}


/***************************** initialise code *****************************/


int initialise( ClassFile* cf, Stack* stp )
{
	int retcode = 0;

	if( cf->static_size )
	{
		if( verbose & 2 )
		{
			char cbuf[255], buf[512];
			sprintClassName( cbuf, cf, cf->this_class );
			sprintf( buf, "Allocating %d words of static data to class \"%s\"\n", (int) cf->static_size, cbuf );
			systemOutPrint( buf, strlen( buf ) );
		}

		cf->static_data = (u4*) javavm_malloc( 4 * cf->static_size );
		memset( cf->static_data, 0, 4 * cf->static_size );
	}

	cf->initialised = 1;

	retcode = executeCLInit( cf, stp );

	return retcode;
}
