/******************************************************************************
 *
 * Filename: memory.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Memory management utilities.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include "config.h"
#include "types.h"
#include "engine.h"
#include "loader.h"
#include "globals.h"

#include "memory.h"


int32 bytes_used = 0;
int32 bytes_peak = 0;
int32 bytes_total = 0;


void* javavm_malloc( size_t size )
{
    void* ret;
    
	bytes_used += size;
	bytes_total += size;

	if( bytes_used > bytes_peak )
		bytes_peak = bytes_used;

	ret = malloc( size );
	
	if( !ret )
	{
	    printf( "FATAL: malloc() failed to allocate %d bytes\n", size );
	    exit( 1 );
	}
	
	return ret;
}


void javavm_free( void* ptr, size_t size )
{
	bytes_used -= size;

	free( ptr );
}


void freeObjectPool()
{
	ObjRefList *list, *prev = NULL;

	for( list = objectPool; list; list = list->next )
	{
		if( prev )
			javavm_free( prev, sizeof(ObjRefList) );

		javavm_free( list->obj->data, 4 * list->obj->cf->object_size );
		javavm_free( list->obj, sizeof(ObjRef) );
		prev = list;
	}

	if( prev )
		javavm_free( prev, sizeof(ObjRefList) );

}


void freeArrayPool()
{
	ObjRefList *list, *prev = NULL;
	int size;

	for( list = arrayPool; list; list = list->next )
	{
		if( prev )
			javavm_free( prev, sizeof(ObjRefList) );

		switch( list->obj->type )
		{
		case T_OBJREF:	size = sizeof(ObjRef*); break;
		case T_BOOLEAN: size = 1; break;
		case T_CHAR:	size = 1; break;
		case T_FLOAT:	size = 4; break;
		case T_DOUBLE:	size = 8; break;
		case T_BYTE:	size = 1; break;
		case T_SHORT:	size = 2; break;
		case T_INT:		size = 4; break;
		case T_LONG:	size = 8; break;
		default: printf( "ERROR: bad array type\n" ); break;
		}

		javavm_free( list->obj->data, size * list->obj->length );
		javavm_free( list->obj, sizeof(ObjRef) );
		prev = list;
	}

	if( prev )
		javavm_free( prev, sizeof(ObjRefList) );

}


void freeClassFiles()
{
	ClassFileRecord *list, *prev = NULL;
	u2 n, m;

	for( list = loadedTable; list; list = list->next )
	{
		if( prev )
			javavm_free( prev, sizeof(ClassFileRecord) );

		/* Constant pool entries */
		for( n = 1; n < list->cf->constant_pool_count; n++ )
		{
			switch( list->cf->constant_pool[n]->tag )
			{
			case CONSTANT_Utf8: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Utf8_info) ); break;
			case CONSTANT_Integer:
			case CONSTANT_Float:
				javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Integer_info) );
				break;
			case CONSTANT_Long:
			case CONSTANT_Double:
				javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Long_info) );
				n++; /* trailing blank entry */
				break;
			case CONSTANT_Class: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Class_info) ); break;
			case CONSTANT_String:
				if( ((CONSTANT_String_info*) list->cf->constant_pool[n])->objref )
				{
					javavm_free( ((CONSTANT_String_info*) list->cf->constant_pool[n])->objref->data, 4 * ((CONSTANT_String_info*) list->cf->constant_pool[n])->objref->cf->object_size );
					javavm_free( ((CONSTANT_String_info*) list->cf->constant_pool[n])->objref, sizeof(ObjRef) );
				}

				javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_String_info) );
				break;
			case CONSTANT_Fieldref: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Fieldref_info) ); break;
			case CONSTANT_Methodref: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_Methodref_info) ); break;
			case CONSTANT_InterfaceMethodref: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_InterfaceMethodref_info) ); break;
			case CONSTANT_NameAndType: javavm_free( list->cf->constant_pool[n], sizeof(CONSTANT_NameAndType_info) ); break;
			default:
				printf( "ERROR: Unknown tag '%d' whilst freeing constant pool\n", list->cf->constant_pool[n]->tag );
				break;
			}
		}

		javavm_free( list->cf->constant_pool, list->cf->constant_pool_count * sizeof(cp_info*) );

		/* Interface entries */
		for( n = 0; n < list->cf->interfaces_count; n++ )
			javavm_free( list->cf->interfaces[n], sizeof(InterfaceEntry) );

		javavm_free( list->cf->interfaces, list->cf->interfaces_count * sizeof(InterfaceEntry*) );

		/* Fields and field attributes */
		for( n = 0; n < list->cf->fields_count; n++ )
		{
			for( m = 0; m < list->cf->fields[n]->attributes_count; m++ )
				javavm_free( list->cf->fields[n]->attributes[m], sizeof(attribute_info) + list->cf->fields[n]->attributes[m]->attribute_length );

			javavm_free( list->cf->fields[n]->attributes, list->cf->fields[n]->attributes_count * sizeof(attribute_info*) );
			javavm_free( list->cf->fields[n], sizeof(field_info) );
		}

		javavm_free( list->cf->fields, list->cf->fields_count * sizeof(field_info*) );

		/* Methods and method attributes */
		for( n = 0; n < list->cf->methods_count; n++ )
		{
			for( m = 0; m < list->cf->methods[n]->attributes_count; m++ )
				javavm_free( list->cf->methods[n]->attributes[m], sizeof(attribute_info) + list->cf->methods[n]->attributes[m]->attribute_length );

			javavm_free( list->cf->methods[n]->attributes, list->cf->methods[n]->attributes_count * sizeof(attribute_info*) );
			javavm_free( list->cf->methods[n], sizeof(method_info) );
		}

		javavm_free( list->cf->methods, list->cf->methods_count * sizeof(method_info*) );

		/* Virtual methods */
		if( list->cf->super || list->cf->interfaces_count )
			javavm_free( list->cf->virtuals, list->cf->methods_count * sizeof(method_info*) );

		/* Class attributes */
		for( n = 0; n < list->cf->attributes_count; n++ )
			javavm_free( list->cf->attributes[n], sizeof(attribute_info) + list->cf->attributes[n]->attribute_length );

		javavm_free( list->cf->attributes, list->cf->attributes_count * sizeof(attribute_info*) );

		/* static data */
		if( list->cf->static_size )
			javavm_free( list->cf->static_data, 4 * list->cf->static_size );

		/* java.lang.Class object */
		javavm_free( list->cf->classobj, sizeof(ObjRef) );

		/* The classfile itself */
		javavm_free( list->cf, sizeof(ClassFile) );

		prev = list;
	}

	if( prev )
		javavm_free( prev, sizeof(ClassFileRecord) );
}
