/******************************************************************************
 *
 * Filename: util.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Class construction utilities.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <string.h>
#include <stdio.h>

#include "config.h"
#include "classfile.h"
#include "engine.h"
#include "loader.h"
#include "linker.h"
#include "disasm.h"
#include "memory.h"
#include "globals.h"

#include "util.h"


#define CLASS_CLASSNAME "java/lang/Class"
#define STRING_CLASSNAME "java/lang/String"
#define THREAD_CLASSNAME "java/lang/Thread"
#define FILEDESC_CLASSNAME "java/io/FileDescriptor"
#define OBJSTRFIELD_CLASSNAME "java/io/ObjectStreamField"


extern void systemOutPrint( const char* buf, int len );


/* Create a new stack frame in intialise it to zeros */
Frame* newStackFrame( method_info* mi, Frame* next )
{
	Frame *frame = NULL;

	if( !mi || !mi->ca || !mi->ca->code )
		return NULL;
	
	frame = javavm_malloc( sizeof( Frame ) + 5 * (mi->ca->max_stack + mi->ca->max_locals) );
	frame->cf = mi->cf;
	frame->mi = mi;
	frame->pc = (u1*) mi->ca->code;
	frame->min_pc = frame->pc;
	frame->max_pc = frame->pc + mi->ca->code_length;
	frame->local = (u4*) &frame->data; /* start of data */
	frame->sp = &frame->local[mi->ca->max_locals]; /* end of locals */
	frame->tlocal = (u1*) &frame->sp[mi->ca->max_stack]; /* end of stack */
	frame->tsp = &frame->tlocal[mi->ca->max_locals]; /* end of tlocals */
	frame->ex = NULL;
	frame->next = next;
	frame->prev = NULL;
	frame->init = NULL;
	frame->stopOnPop = 0;

	memset( frame->local, 0, 5 * (mi->ca->max_stack + mi->ca->max_locals) );

	return frame;
}


/* Free a stack frame created with newStackFrame() */
void freeStackFrame( Frame* fp, method_info* mi )
{
	javavm_free( fp, sizeof( Frame ) + 5 * (mi->ca->max_stack + mi->ca->max_locals) );
}


/* Create an instance of java.lang.Class */
ObjRef* newJavaLangClass( ClassFile* cf )
{
	ClassFile* classcf;
	ObjRef* classobj;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

    if( !(classcf = getClass( &stack, CLASS_CLASSNAME )) )
		return NULL; /* ###? exception */

	classobj = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	classobj->cf = classcf;
	classobj->data = (u4*) cf;  /* Big cheat - there is no data for Class, so we use ours! */
	classobj->type = 0;
	classobj->length = 1;

	return classobj;
}


/* Create an instance of java.lang.String */
ObjRef* newJavaLangString( u1* bytes, int length )
{
	ClassFile* stringcf;
	javaLangString *data;
	ObjRef* value;
	ObjRef* objref;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(stringcf = getClass( &stack, STRING_CLASSNAME )) )
	{
	    if( verbose ) printf( "ERROR: newJavaLangString: class java.lang.String not loaded\n" );
		return NULL; /* ### exception */
    }

	value = newObjectRefAndArray( T_BYTE, length, 1 );
	memcpy( value->data, bytes, length );

	objref = newObjectRefAndObject( stringcf );
	data = (javaLangString*) objref->data;
	data->value = (u4) value;
	data->offset = 0;
	data->count = length;
	return objref;
}


/* Create an instance of java.lang.Thread */
ObjRef* newJavaLangThread( const char* threadName, ObjRef* group )
{
	ClassFile* threadcf;
	javaLangThread *data;
	ObjRef* name;
	ObjRef* objref;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(threadcf = getClass( &stack, THREAD_CLASSNAME )) )
	{
	    if( verbose ) printf( "ERROR: newJavaLangString: class java.lang.Thread not loaded\n" );
		return NULL; /* ### exception */
    }

	name = newObjectRefAndArray( T_BYTE, strlen(threadName), 1 );
	memcpy( name->data, threadName, strlen(threadName) );

	objref = newObjectRefAndObject( threadcf );
	data = (javaLangThread*) objref->data;
	data->name = (u4) name;
	data->priority = 5; /* NORM */
	data->threadQ = (u4) NULL;
	data->PrivateInfo = 0;
	data->eetop = 0;
	data->single_step = 0; /* false */
	data->daemon = 0; /* false */
	data->stillborn = 0; /* false */
	data->target = (u4) NULL;
	data->group = (u4) group;
	data->initial_stack_memory = 0;
	return objref;
}


ObjRef* newJavaIoFileDescriptor( int fd )
{
	ClassFile* filedesccf;
	ObjRef* objref;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(filedesccf = getClass( &stack, FILEDESC_CLASSNAME )) )
		return NULL; /* ### exception */

	objref = newObjectRefAndObject( filedesccf );
	objref->data[0] = (u4) fd;

	return objref;
}


ObjRef* newJavaIoObjectStreamField( ObjRef* name, u1 type, u4 offset, ObjRef* typeString )
{
	ClassFile* objstrfieldcf;
	ObjRef* objref;
	javaIoObjectStreamField* data;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	if( !(objstrfieldcf = getClass( &stack, OBJSTRFIELD_CLASSNAME )) )
		return NULL; /* ### exception */

	objref = newObjectRefAndObject( objstrfieldcf );
	data = (javaIoObjectStreamField*) objref->data;
	data->name = (u4) name;
	data->type = type;
	data->offset = offset;
	data->typeString = (u4) typeString;

	return objref;
}


/* Creates a new object reference which points to a blank object of the given class */
ObjRef* newObjectRefAndObject( ClassFile* cf )
{
	ObjRef* objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	objref->cf = cf;
	objref->type = 0;
	objref->data = (u4*) javavm_malloc( 4 * cf->object_size );
	objref->length = 1;
	memset( objref->data, 0, 4 * cf->object_size );
	return objref;
}


/* Creates a new object reference which points to a blank array of the given type */
ObjRef* newObjectRefAndArray( u1 type, int32 length, u4 size )
{
	ObjRef* objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );
	objref->cf = NULL;
	objref->type = type;
	objref->data = NULL;
	objref->length = length;

	if( (length > 0) && (size > 0) )
	{
		objref->data = javavm_malloc( size * length );
		memset( objref->data, 0, size * length );
	}
	return objref;
}


ClassFile* getClass( Stack *stp, const char* classname )
{
    ClassFile *cf;
    
    cf = checkLoaded( classname );
    
	if( !cf )
		cf = load( classname, stp );

	return cf;
}


int checkType( ClassFile* cf, const char* classname )
{
	ClassFile* checkcf;
	Stack stack;

	stack.ex = NULL;
	stack.fp = NULL;

	checkcf = getClass( &stack, classname );

	if( !checkcf )
		return 0;

	return (checkcf == cf);
}


char* extractFromJavaLangString( char* buf, int buflen, ObjRef* str )
{
	if( str->cf && checkType( str->cf, STRING_CLASSNAME ) && str->data && str->data[0] )
	{
		javaLangString* data = (javaLangString*) str->data;
		ObjRef* array = (ObjRef*) data->value;

		if( array && !array->cf
			&& ((array->type == T_BYTE) || (array->type == T_CHAR)) )
		{
			if( (data->offset <= array->length)
				&& (data->count <= array->length)
				&& ((int) data->count <= buflen) )
			{
				u1* src = (u1*) array->data;
				strncpy( buf, (const char*) &src[data->offset], data->count );
				buf[data->count] = '\0';
				return buf;
			}
		}
	}

	return NULL;
}


int newUnsatisfiedLinkError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintMethodRef( buf, stp->fp->cf, index );
	printf( "ERROR: No Java code or no native implementation for method %s\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, UnsatisfiedLinkError, buf );
}


int newUnsatisfiedLinkErrorIface( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintInterfaceMethodRef( buf, stp->fp->cf, index );
	printf( "ERROR: No Java code or no native implementation for method %s\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, UnsatisfiedLinkError, buf );
}


int newNoSuchMethodError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintMethodRef( buf, stp->fp->cf, index );
	printf( "ERROR: Method %s not linked\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchMethodError, buf );
}


int newNoSuchMethodErrorIface( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintInterfaceMethodRef( buf, stp->fp->cf, index );
	printf( "ERROR: Method %s not linked\n", buf );
	
	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchMethodError, buf );
}


int newNoSuchFieldError( Stack* stp, ClassFile* cf, u2 index )
{
	char buf[255];
	sprintFieldRef( buf, cf, index );

	if( verbose )
		whereAreWe( stp->fp );

	return newCompleteException( stp, NoSuchFieldError, buf );
}


const char* exceptionKindToString( ExceptionKind kind )
{
	switch( kind )
	{
	case VerifyError:            return "java/lang/VerifyError";
	case NoClassDefFoundError:   return "java/lang/NoClassDefFoundError";
	case IllegalAccessError:     return "java/lang/IllegalAccessError";
	case ClassFormatError:       return "java/lang/ClassFormatError";
	case NoSuchFieldError:       return "java/lang/NoSuchFieldError";
	case NoSuchMethodError:      return "java/lang/NoSuchMethodError";
	case InternalError:          return "java/lang/InternalError";
	case UnsatisfiedLinkError:	 return "java/lang/UnsatisfiedLinkError";

	case NullPointerException:   return "java/lang/NullPointerException";
	case ArrayIndexOutOfBoundsException: return "java/lang/ArrayIndexOutOfBoundsException";
	case ArithmeticException:    return "java/lang/ArithmeticException";
	case ClassCastException:     return "java/lang/ClassCastException";
	case ClassNotFoundException: return "java/lang/ClassNotFoundException";
	case IOException:            return "java/io/IOException";
	case NegativeArraySizeException: return "java/lang/NegativeArraySizeException";

	case JavaException:			 return NULL;
	}

	return NULL;
}


void whereAreWe( Frame* baseFrame )
{
	Frame* fp;
	char buffer[255];
	char *buf;
	int len;

	systemOutPrint( "Where are we:\n", 14 );

	if( !baseFrame )
	{
		systemOutPrint( "    null stack pointer!\n", 24 );
		return;
	}

	/* Add each frames */
	for( fp = baseFrame; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buffer, "    at " ); buf = buffer + len;

		if( !fp->mi->cf )
		{
			systemOutPrint( buffer, strlen( buffer ) );
			systemOutPrint( "\n", 1 );
			continue;
		}

		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		if( fp->mi->lnta )
		{
			int pc = fp->pc - fp->min_pc;
			int best = 0;
			int bestn = fp->mi->lnta->line_number_table_length;
			int n;

			for( n = 0; n < fp->mi->lnta->line_number_table_length; n++ )
			{
				if( (pc >= fp->mi->lnta->line_number_table[n]->start_pc)
					&& (fp->mi->lnta->line_number_table[n]->start_pc > best) )
				{
					best = fp->mi->lnta->line_number_table[n]->start_pc;
					bestn = n;
				}
			}

			if( bestn < fp->mi->lnta->line_number_table_length )
			{
				len = sprintf( buf, ":%d", fp->mi->lnta->line_number_table[bestn]->line_number ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		len = sprintf( buf, "#%d", (int) (fp->pc - fp->min_pc) ); buf += len;

		systemOutPrint( buffer, strlen( buffer ) );
		systemOutPrint( ")\n", 2 );
	}
}


ObjRef* createMultiObjrefArray( int dimension, int32* length, int size, u1 type )
{
	ObjRef *objref;
	int n;
	
	/*printf( "createMultiObjrefArray( dimension = %d, length[0] = %d, size = %d, type = %d )\n",
		dimension, length[0], size, type );*/

	objref = (ObjRef*) javavm_malloc( sizeof( ObjRef ) );

	/*printf( "objref = 0x%08x\n", objref );*/

	objref->cf = NULL;
	objref->type = (dimension == 1) ? type : T_OBJREF;
	objref->length = length[0];

	/* ### Should be T_ARRAY and should check constant pool entry */

	if( length[0] > 0 )
	{
		if( dimension == 1 )
		{
			objref->data = javavm_malloc( size * length[0] );
			memset( objref->data, 0, size * length[0] );
			addToList( objref, &arrayPool );
		}
		else
		{
			ObjRef** elements;

			objref->data = javavm_malloc( sizeof( ObjRef* ) * length[0] );
			memset( objref->data, 0, sizeof( ObjRef* ) * length[0] );

			elements = (ObjRef**) objref->data;

			for( n = 0; n < length[0]; n++ )
				elements[n] = createMultiObjrefArray( dimension-1, &length[1], size, type );
		}
	}
	else
	{
		objref->data = NULL;
	}

	return objref;
}


/* add an element to a object reference list */
void addToList( ObjRef* objref, ObjRefList** pool )
{
	ObjRefList* list = (ObjRefList*) javavm_malloc( sizeof( ObjRefList ) );
	list->obj = objref;
	list->next = *pool;
	*pool = list;
}


int findFreeFD()
{
	int n;

	for( n = 3; (n < 255) && fdList[n]; n++ );

	if( n == 255 )
	{
		printf( "ERROR: Out of file descriptors\n" );
		return -1;
	}

	if( n > fdLast )
		fdLast = n;

	return n;
}
