/******************************************************************************
 *
 * Filename: main.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Main program for virtual machine.
 *
 * (c) Copyright R.H.Thornburrow, 1998
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef WIN32
#include <sys/timeb.h>
#else
#include <sys/time.h>
#endif /* WIN32 */

#include <time.h>
#include <signal.h>

#include "../engine/config.h"
#include "../engine/classfile.h"
#include "../engine/loader.h"
#include "../engine/linker.h"
#include "../engine/disasm.h"
#include "../engine/engine.h"
#include "../engine/util.h"
#include "../engine/instructions.h"
#include "../engine/memory.h"
#include "../engine/globals.h"


int verbose = 0;
int instruction_debug = 0;
int method_debug = 0;
int field_debug = 0;
int arithmetic_debug = 0;
int instruction_count = 0;
#ifdef _DEBUG
int instruction_count_dump = 0;
#endif

Stack stack;
ObjRef* threads[1];


/* fwd ref */
void dumpStack( Stack* stp );


#ifdef SUPPORTS_SIGNALS
void signal_handler( int sig )
{
	printf( "SIGSEGV detected - full thread dump:\n" );
	printf( "    \"main\":\n" );

	dumpStack( &stack );

	exit( 0 );
}
#endif /* SUPPORTS_SIGNALS */


void dumpStack( Stack* stp )
{
	Frame *fp;
	char buffer[255];
	char *buf;
	int len;

	/* Add each stack frame */
	for( fp = stp->fp; fp; fp = (fp->next ? fp->next : fp->init) )
	{
		len = sprintf( buffer, "        at " ); buf = buffer + len;

		if( !fp->mi->cf )
		{
			printf( "%s\n", buffer );
			continue;
		}

		len = sprintClassName( buf, fp->mi->cf, fp->mi->cf->this_class ); buf += len;
		len = sprintf( buf, "." ); buf += len;

		if( CHECKCONSTANT( fp->mi->cf, fp->mi->name_index, CONSTANT_Utf8 ) )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->name_index ); buf += len;
		}

		len = sprintf( buf, "(" ); buf += len;

		if( fp->mi->cf->sfa )
		{
			len = sprintSimpleName( buf, fp->mi->cf, fp->mi->cf->sfa->sourcefile_index ); buf += len;
		}

		if( fp->mi->lnta )
		{
			int pc = fp->pc - fp->min_pc;
			int best = 0;
			int bestn = fp->mi->lnta->line_number_table_length;
			int n;

			for( n = 0; n < fp->mi->lnta->line_number_table_length; n++ )
			{
				if( (pc >= fp->mi->lnta->line_number_table[n]->start_pc)
					&& (fp->mi->lnta->line_number_table[n]->start_pc > best) )
				{
					best = fp->mi->lnta->line_number_table[n]->start_pc;
					bestn = n;
				}
			}

			if( bestn < fp->mi->lnta->line_number_table_length )
			{
				len = sprintf( buf, ":%d", fp->mi->lnta->line_number_table[bestn]->line_number ); buf += len;
			}
		}
		
		/* extra position info giving exact pc of exception */
		len = sprintf( buf, "#%d", (int) (fp->pc - fp->min_pc) ); buf += len;

		printf( "%s)\n", buffer );
	}
}


void reportException( Stack* stp )
{
	Exception *ex;

	if( stp->fp && stp->fp->ex )
		ex = stp->fp->ex;
	else
		ex = stp->ex;

	printf( "Unhandled exception caught by JavaVM:\n" );

	if( !ex )
	{
		printf( "\tUnknown Exception\n" );
		return;
	}

	if( ex->kind == JavaException )
	{
		if( ex->ex && ex->ex->cf )
		{
			char buf[1024];
			sprintClassName( buf, ex->ex->cf, ex->ex->cf->this_class );
			printf( "%s", buf );

			if( ex->ex->data && ex->ex->data[1] )
			{
				ObjRef* string = (ObjRef*) ex->ex->data[1];

				if( string && string->cf && checkType( string->cf, "java/lang/String" ) && string->data && string->data[0] )
				{
					ObjRef* array = (ObjRef*) string->data[0];

					if( array && !array->cf && ((array->type == T_BYTE) || (array->type == T_CHAR)) )
					{
						char buf[1024];
						strncpy( buf, (const char*) array->data, array->length );
						buf[array->length] = '\0';
						printf( ": %s", buf );
					}
				}
			}

			printf( "\n" );

			/* stack trace */
			if( ex->ex->data && ex->ex->data[0] )
				printf( "%s", (char*) ex->ex->data[0] );
		}
		else
		{
			printf( "\tUnknown Java Exception\n" );
			dumpStack( stp );
		}

		return;
	}

	if( exceptionKindToString( ex->kind ) )
		printf( "\t%s", exceptionKindToString( ex->kind ) );
	else
		printf( "\tUnknown Exception" );

	if( ex->message )
		printf( ": %s", ex->message );

	printf( "\n" );
	dumpStack( stp );
}


int vmInit()
{
	ClassFile *systemClass, *threadClass, *threadGroupClass;
	ObjRef* systemThreadGroup = NULL;
	int n, retcode = 0;

	/* init file descriptor list to 0 */
	for( n = 0; n < 255; n++ )
		fdList[n] = 0;

	/* init primitive class objects to null */
	for( n = 0; n < T_LAST; n++ )
		primitiveClass[n] = NULL;

	/* load and init java.lang.System */
	systemClass = load( "java/lang/System", &stack );

	if( !stack.ex && systemClass )
	{
		retcode = forceExecuteStaticJavaMethod( systemClass, &stack, "initializeSystemClass", "()V" );

		while( !retcode )
			retcode = executeInstruction( &stack );
	}

	if( !systemClass || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}


	/* create "system" ThreadGroup object */
	threadGroupClass = load( "java/lang/ThreadGroup", &stack );

	if( !stack.ex && threadGroupClass )
	{
		/* create new ThreadGroup object */
		systemThreadGroup = newObjectRefAndObject( threadGroupClass );

		/* call constructor ThreadGroup.<init>() */
		retcode = forceExecuteJavaMethod( systemThreadGroup, &stack, "<init>", "()V" );

		while( !retcode )
			retcode = executeInstruction( &stack );
	}

	if( !threadGroupClass || !systemThreadGroup || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}


	/* Create "main" thread object */
	threadClass = load( "java/lang/Thread", &stack );

	if( !stack.ex && threadClass )
	{
		/* create a new Thread object */
		threads[0] = newJavaLangThread( "main", systemThreadGroup );
	}

	if( !threadClass || stack.ex || (retcode == 2) )
	{
		reportException( &stack );
		return 2;
	}

	return 0;
}


void showUsage()
{
	printf( "Taurus Java Virtual Machine %s v%s %s\n", JAVAVM_PLATFORM, JAVAVM_VERSION, JAVAVM_RELEASE );
	printf( "(C) Copyright 2001, Taurus Software\n\n" );

	printf( "usage: javavm [-options] <class>\n\n" );
	printf( "where options include:\n" );
	printf( "    -[h]elp             this message\n" );
#ifdef _DEBUG
	printf( "    -[d]isasm <level>   debug mode\n" );
#endif
	printf( "    -[i]trace           instruction trace\n" );
	printf( "    -[m]trace           method trace\n" );
	printf( "    -[f]trace           field trace\n" );
#ifdef _DEBUG
	printf( "    -[c]ount            instruction count dump\n" );
#endif
	printf( "    -[a]trace           arithmetic trace\n" );
	printf( "    -[v]erbose          verbose mode\n" );
	printf( "    -version            prints version number\n" );
}


int main( int argc, char* argv[] )
{
#ifdef _DEBUG
	int level = 0;
#endif
	ClassFile *ptr = NULL;
	int retcode = 0;
#ifdef WIN32
	struct _timeb startbuf, endbuf;
#else
	/*struct timeb startbuf, endbuf;*/
	struct timeval startbuf, endbuf;
#endif

	if( argc <= 1 )
	{
		showUsage();
		return 1;
	}

	argv++; argc--;
	while( *argv && ((*argv)[0] == '-') )
	{
		switch( (*argv)[1] )
		{
		case 'a':
			if( (strcmp( *argv, "-a" ) == 0) || (strcmp( *argv, "-atrace" ) == 0) ) {
				arithmetic_debug = 1;
			} else {
				showUsage();
				return 1;
			}
			break;
		case 'c':
#ifdef _DEBUG
			if( (strcmp( *argv, "-c" ) == 0) || (strcmp( *argv, "-count" ) == 0) ) {
				instruction_count_dump = 1;
			} else
#endif /* _DEBUG */
			{
				showUsage();
				return 1;
			}
			break;
		case 'd':
#ifdef _DEBUG
			if( (strcmp( *argv, "-d" ) == 0)
				|| (strcmp( *argv, "-disasm" ) == 0) )
			{
				argv++; argc--;

				if( *argv && (*argv[0] != '-') )
				{
					level = atoi( *argv );
				}
				else
				{
					printf( "usage: javavm [-disasm <level>] <class>\n\n" );
					printf( "where level is a combination of:\n" );
					printf( "    1 = method code\n" );
					printf( "    2 = constant pool table\n" );
					printf( "    4 = class attributes\n" );
					printf( "    8 = method attributes\n" );
					return 1;
				}
			}
			else
#endif /* _DEBUG */
			{
				showUsage();
				return 1;
			}
			break;
		case 'i':
			if( (strcmp( *argv, "-i" ) == 0) || (strcmp( *argv, "-itrace" ) == 0) ) {
				instruction_debug = 1;
			} else {
				showUsage();
				return 1;
			}
			break;
		case 'f':
			if( (strcmp( *argv, "-f" ) == 0) || (strcmp( *argv, "-ftrace" ) == 0) ) {
				field_debug = 1;
			} else {
				showUsage();
				return 1;
			}
			break;
		case 'm':
			if( (strcmp( *argv, "-m" ) == 0) || (strcmp( *argv, "-mtrace" ) == 0) ) {
				method_debug = 1;
			} else {
				showUsage();
				return 1;
			}
			break;
		case 'v':
			if( (strcmp( *argv, "-v" ) == 0) || (strcmp( *argv, "-verbose" ) == 0) ) {
				verbose++;
			} else if( strcmp( *argv, "-version" ) == 0 ) {
				printf( "java version \"1.1\"\n" );
				printf( "Taurus Java Virtual Machine %s v%s %s\n", JAVAVM_PLATFORM, JAVAVM_VERSION, JAVAVM_RELEASE );
				printf( "(C) Copyright 2001, Taurus Software\n" );
				return 0;
			} else {
				showUsage();
				return 1;
			}
			break;
		case 'h':
		default:
			showUsage();
			return 1;
		}
		argv++; argc--;
	}

	if( !*argv )
	{
		showUsage();
		return 1;
	}

#ifdef _DEBUG
	if( level > 0 )
	{
		char filename[255];
		Stack stack;

		strcpy( filename, *argv );
		strcpy( &filename[strlen(*argv)], ".class" );
		filename[strlen(*argv) + 6] = '\0';

		stack.retcode = 0;
		ptr = readClassFile( filename, filename, &stack );

		if( ptr )
			disassemble( ptr, level );

		printf( "\nPress return to exit.\n" );
		fgetc( stdin );
		return 0;
	}
#endif /* _DEBUG */

#ifdef WIN32
	_ftime( &startbuf );
#else
	gettimeofday( &startbuf, NULL );
#endif

#ifdef SUPPORTS_SIGNALS
	if( signal( SIGSEGV, signal_handler ) == SIG_ERR )
	{
		printf( "Failed to attach signal handler\n" );
		return 1;
	}
#endif /* SUPPORTS_SIGNALS */


	stack.ex = NULL;
	stack.fp = NULL;
	stack.tail = NULL;

	if( vmInit() == 2 )
		return -1;

	if( !stack.ex )
		ptr = load( *argv, &stack );

	if( !ptr )
	{
		reportException( &stack );
		return -1;
	}

	argv++; argc--;
	retcode = execute( ptr, &stack, argc, argv );

#ifdef WIN32
	_ftime( &endbuf );
#else
	gettimeofday( &endbuf, NULL );
#endif

	if( 1 )
	{
		double elapsed, ips;
#ifdef WIN32
		double starttime = startbuf.time;
		double endtime = endbuf.time;

		starttime *= 1000;
		starttime += startbuf.millitm;
		endtime *= 1000;
		endtime += endbuf.millitm;
		elapsed = (endtime - starttime) / 1000;
#else
		double starttime = startbuf.tv_sec;
		double endtime = endbuf.tv_sec;

		starttime *= 1000000;
		starttime += startbuf.tv_usec;
		endtime *= 1000000;
		endtime += endbuf.tv_usec;
		elapsed = (endtime - starttime) / 1000000;
#endif
		ips = instruction_count / elapsed;
		printf( "Executed %d instructions in %.3f seconds (%.2f instructions/sec)\n", instruction_count, elapsed, ips );
	}

	if( verbose > 1 )
	{
		if( objectPool )
		{
			ObjRefList *item, *prev, *next;
			ClassFile* type;
			int objCount = 0;

			for( item = objectPool; item; item = item->next )
				objCount++;

			printf( "Allocated total of %d objects:\n", objCount );

			while( objectPool )
			{
				type = objectPool->obj->cf;
				objCount = 1;

				prev = objectPool;
				item = objectPool->next;

				while( item )
				{
					if( item->obj->cf == type )
					{
						objCount++;
						prev->next = item->next;
						javavm_free( item, sizeof(ObjRefList) );
						item = prev->next;
						continue;
					}
					
					prev = item;
					item = item->next;
				} /* item */

				printf( "\t%d %s\n", objCount, type->name );

				next = objectPool->next;
				javavm_free( objectPool, sizeof(ObjRefList) );
				objectPool = next;
			} /* objectPool */
		} /* objectPool */

		if( arrayPool )
		{
			ObjRefList *item, *prev, *next;
			ClassFile* ctype;
			int btype;
			int objCount = 0;

			for( item = arrayPool; item; item = item->next )
				objCount++;

			printf( "Allocated total of %d arrays:\n", objCount );

			while( arrayPool )
			{
				ctype = arrayPool->obj->cf;
				btype = arrayPool->obj->type;
				objCount = 1;

				prev = arrayPool;
				item = arrayPool->next;

				while( item )
				{
					if( (item->obj->cf == ctype) && (item->obj->type == btype) )
					{
						objCount++;
						prev->next = item->next;
						javavm_free( item, sizeof(ObjRefList) );
						item = prev->next;
						continue;
					}
					
					prev = item;
					item = item->next;
				} // item

				if( ctype )
				{
					printf( "\t%d %s[]\n", objCount, ctype->name );
				}
				else
				{
					switch( btype )
					{
					case T_BOOLEAN: printf( "\t%d boolean[]\n", objCount ); break;
					case T_CHAR:	printf( "\t%d char[]\n",    objCount ); break;
					case T_FLOAT:	printf( "\t%d float[]\n",   objCount ); break;
					case T_DOUBLE:	printf( "\t%d double[]\n",  objCount ); break;
					case T_BYTE:	printf( "\t%d byte[]\n",    objCount ); break;
					case T_SHORT:	printf( "\t%d short[]\n",   objCount ); break;
					case T_INT:		printf( "\t%d int[]\n",     objCount); break;
					case T_LONG:	printf( "\t%d long[]\n",    objCount ); break;
					}
				}

				next = arrayPool->next;
				javavm_free( arrayPool, sizeof(ObjRefList) );
				arrayPool = next;
			} /* while */
		} /* if */
	} /* verbose */

#ifdef _DEBUG
	if( instruction_count_dump )
	{
		int inst;

		printf( "Instruction counts:\n" );

		for( inst = 0; inst < 256; inst++ )
		{
			if( (instructionTable[inst].name != NULL)
				&& (instructionTable[inst].count == 0)
				/*(instructionTable[inst].count > 0)
				&& (instructionTable[inst].count <= 5)*/ )
				printf( "\t%s\t\t%d\n", instructionTable[inst].name, ((int) instructionTable[inst].count) );
		}
	}
#endif

	if( retcode == 2 /* exception */ )
		reportException( &stack );

	freeObjectPool();
	freeArrayPool();
	freeClassFiles();

	if( 1 )
	{
		printf( "Memory total %d kbytes, peak %d kbytes, used %d kbytes\n",
			(int) (bytes_total / 1024), (int) (bytes_peak / 1024), (int) (bytes_used / 1024) );
	}

	/*fgetc( stdin );*/

	return ((retcode == 2) ? -1 : 0);
}
