/******************************************************************************
 *
 * Filename: java_io_File.c
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  19th January 1998
 *
 * Purpose:  Native method implementations for java.io.File.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <sys/stat.h>    /* for file stat stuff */

#ifdef WIN32
#include <io.h>          /* for file find stuff */
#else
#include <dirent.h>
#endif

#include "../engine/config.h"

#include "java_io_File.h"

#include "../engine/util.h"
#include "../engine/globals.h"

#ifdef WIN32

#define STAT _stat
#define IS_DIRECTORY(buf)	((buf.st_mode & _S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & _S_IFREG) ? 1 : 0)
#define MODIFIED_TIME(buf)	buf.st_mtime
#define CAN_READ(buf)		((buf.st_mode & _S_IREAD) ? 1 : 0)

#else

#define STAT stat
#define IS_DIRECTORY(buf)	((buf.st_mode & S_IFDIR) ? 1 : 0)
#define IS_FILE(buf)		((buf.st_mode & S_IFREG) ? 1 : 0)
#define MODIFIED_TIME(buf)	buf.st_mtime
#define CAN_READ(buf)		((buf.st_mode & S_IREAD) ? 1 : 0)

#endif /* WIN32 */


/* boolean exists0(); */
jboolean Java_java_io_File_canRead0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		struct STAT buf;
		int ret = STAT( filename, &buf );

		if( ret == 0 )
			return CAN_READ(buf);

		/*printf( "File.canRead0( '%s' ) returning false.\n", filename );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/* boolean delete0() */
jboolean Java_java_io_File_delete0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		printf( "WARNING: java.io.File.delete0() for '%s' dummied to true.\n", filename );
		return 1;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/* boolean exists0(); */
jboolean Java_java_io_File_exists0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		struct STAT buf;
		int ret = STAT( filename, &buf );
		return ( (ret == 0) ? 1 : 0 );
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/**
 * Tests if the file represented by this <code>File</code> object is an
 * absolute pathname. The definition of an absolute pathname is system 
 * dependent. For example, on UNIX, a pathname is absolute if its 
 * first character is the separator character. On Windows platforms, 
 * a pathname is absolute if its first character is an ASCII '&#92;' or 
 * '/', or if it begins with a letter followed by a colon. 
 *
 * @return  <code>true</code> if the pathname indicated by the
 *          <code>File</code> object is an absolute pathname;
 *          <code>false</code> otherwise.
 * @see     java.io.File#getPath()
 * @see     java.io.File#separator
 * @since   JDK1.0
 *
 * public native boolean isAbsolute();
 */
jboolean Java_java_io_File_isAbsolute( JNIEnv* env, jobject self )
{
	printf( "WARNING: File.isAbsolute() dummied.\n" );
	return 1;
}


/* boolean isDirectory0(); */
jboolean Java_java_io_File_isDirectory0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		struct STAT buf;
		int ret;

#ifdef WIN32
		if( (strlen(filename) > 1) && (filename[ strlen(filename) - 1 ] == '\\') )
			filename[ strlen(filename) - 1 ] = '\0';
#endif

		ret = STAT( filename, &buf );

		if( ret == 0 )
			return IS_DIRECTORY(buf);

		/*printf( "java.io.File.isDirectory0() for '%s' returning false.\n", filename );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/* boolean isFile0(); */
jboolean Java_java_io_File_isFile0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		struct STAT buf;
		int ret = STAT( filename, &buf );

		if( ret == 0 )
			return IS_FILE(buf);

		/*printf( "java.io.File.isFile0() for '%s' returning false.\n", filename );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/* long lastModified0(); */
jlong Java_java_io_File_lastModified0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		struct STAT buf;
		int ret = STAT( filename, &buf );

		if( ret == 0 )
			return MODIFIED_TIME(buf);

		/*printf( "java.io.File.lastModified0() for '%s' returning to 0.\n", filename );*/
		return 0;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}


/* String[] list0(); */
jobject Java_java_io_File_list0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		ObjRef * array;
		int count = 0;
#ifdef WIN32
		struct _finddata_t c_file;
		long hFile;

		strcat( filename, "\\*.*" );

		/* Find all files in given directory */
		if( (hFile = _findfirst( filename, &c_file )) != -1L )
		{
			count++;

			/* Find the rest of the files */
			while( _findnext( hFile, &c_file ) == 0 )
				count++;

		   _findclose( hFile );
		}
#else
		DIR *dir;
		struct dirent *entry;

		if( (dir = opendir( filename )) )
		{
			while( (entry = readdir( dir )) )
			count++;
		}

		closedir( dir );
#endif
		array = newObjectRefAndArray( T_OBJREF, count, sizeof(ObjRef*) );
		addToList( array, &arrayPool );

		if( count > 0 )
		{
			count = 0;
#ifdef WIN32
			if( (hFile = _findfirst( filename, &c_file )) != -1L )
			{
				array->data[count] = (u4) newJavaLangString( (u1*) c_file.name, strlen( c_file.name ) );
				addToList( (ObjRef*) array->data[count], &objectPool );
				count++;

				/* Find the rest of the files */
				while( _findnext( hFile, &c_file ) == 0 )
				{
					array->data[count] = (u4) newJavaLangString( (u1*) c_file.name, strlen( c_file.name ) );
					count++;
				}

			   _findclose( hFile );
			}
#else
			if( (dir = opendir( filename )) )
			{
				while( (entry = readdir( dir )) )
				{
				    array->data[count] = (u4) newJavaLangString( (u1*) entry->d_name, strlen( entry->d_name ) );
					addToList( (ObjRef*) array->data[count], &objectPool );
					count++;
				}
			}

			closedir( dir );
#endif
		}

		return array;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return NULL;
}


/* boolean mkdir0() */
jboolean Java_java_io_File_mkdir0( JNIEnv* env, jobject self )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, (ObjRef*) self->data[0] ) )
	{
		printf( "WARNING: java.io.File.mkdir0() for '%s' dummied to true.\n", filename );
		return 1;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to get path from java.io.File" );
	return 0;
}
