/******************************************************************************
 *
 * Filename: java_io_FileOutputStream.h
 *
 * Author:   Robert Thornburrow
 * Project:  JavaVM
 * Created:  25th March 2002
 *
 * Purpose:  Native method implementations for java.io.FileOutputStream.
 *
 * (c) Copyright R.H.Thornburrow, 2002
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef WIN32
#include <errno.h>
#endif

#include "../engine/config.h"

#include "java_io_FileOutputStream.h"

#include "../engine/native-io.h"
#include "../engine/util.h"
#include "../engine/globals.h"
#include "../engine/memory.h"


/******************************************************************************
 *
 * class java.io.FileOutputStream
 *
 *
 * The system dependent file descriptor. The value is
 * 1 more than actual file descriptor. This means that
 * the default value 0 indicates that the file is not open.
 *
 * class FileOutputStream
 * {
 *		private FileDescriptor fd;
 * };
 *
 *****************************************************************************/


/**
 * Opens a file, with the specified name, for writing.
 * @param name name of file to be opened
 *
 * private native void open(String name) throws IOException;
 */
void Java_java_io_FileOutputStream_open( JNIEnv* env, jobject self, jobject name )
{
	char filename[255];

	if( extractFromJavaLangString( filename, 255, name ) )
	{
		int fd = findFreeFD();

		if( !(fdList[fd] = fopen( filename, "wb" )) )
		{
			char buf[1024];
#ifdef WIN32
			sprintf( buf, "Failed to open file for writing \"%s\"", filename );
#else
			sprintf( buf, "Failed to open file for writing \"%s\" (errno = %d)", filename, errno );
#endif
			env->retcode = newCompleteException( env, IOException, buf );
			return;
		}

		self->data[0] = (u4) newJavaIoFileDescriptor( fd + 1 );
		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Can't get value of java.lang.String" );
}


/**
 * Opens a file, with the specified name, for appending.
 * @param name name of file to be opened
 *
 * private native void openAppend(String name) throws IOException;
 */
void Java_java_io_FileOutputStream_openAppend( JNIEnv* env, jobject self, jobject name )
{
	/* ### Implement me! */
	env->retcode = newCompleteException( env, InternalError, "FileOutputStream.openAppend() not implemented!" );
	return;
}


/**
 * Writes the specified byte to this file output stream. 
 *
 * @param      b   the byte to be written.
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * public native void write(int b) throws IOException;
 */
void Java_java_io_FileOutputStream_write( JNIEnv* env, jobject self, jint b )
{
	ObjRef *fileDesc;
	u4 fd;
	char buf[2];

	fileDesc = (ObjRef*) self->data[0];

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd - 1 )
	{
	case 0: /* stdin */
		env->retcode = newCompleteException( env, IOException, "Write operation not allowed on input stream" );
		return;
	case 1: /* stdout */
	case 2: /* stderr */
		sprintf( buf, "%c", (u2) b );
		systemOutPrint( buf, 1 );
		break;
	default:
		fputc( b, fdList[fd-1] );
		break;
	}
}


/**
 * Writes a sub array as a sequence of bytes.
 * @param b the data to be written
 * @param off the start offset in the data
 * @param len the number of bytes that are written
 * @exception IOException If an I/O error has occurred.
 *
 * private native void writeBytes(byte b[], int off, int len) throws IOException;
 */
void Java_java_io_FileOutputStream_writeBytes( JNIEnv* env, jobject self, jobject array, jint off, jint len )
{
	ObjRef *fileDesc;
	int32 n;
	u4 fd;

	fileDesc = (ObjRef*) self->data[0];

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	if( array && (array->cf == NULL) && (array->type == T_BYTE) )
	{
		const char* src = (const char*) array->data;

		switch( fd - 1 )
		{
		case 0: /* stdin */
			env->retcode = newCompleteException( env, IOException, "Write operation not allowed on input stream" );
			return;
		case 1: /* stdout */
		case 2: /* stderr */
			{
			char* buf = (char*) javavm_malloc( len + 1 );
			strncpy( buf, src + off, len );
			buf[ len ] = '\0';
			systemOutPrint( buf, len );
			javavm_free( buf, len+1 );
			}
			break;
		default:
			for( n = 0; n < len; n++ )
				fputc( src[n], fdList[fd-1] );
			break;
		}

		return;
	}

	env->retcode = newCompleteException( env, InternalError, "Unable to write bytes to FileOutputStream" );
}


/**
 * Closes this file output stream and releases any system resources 
 * associated with this stream. 
 *
 * @exception  IOException  if an I/O error occurs.
 * @since      JDK1.0
 *
 * public native void close() throws IOException;
 */
void Java_java_io_FileOutputStream_close( JNIEnv* env, jobject self )
{
	ObjRef *fileDesc;
	u4 fd;

	fileDesc = (ObjRef*) self->data[0];

	if( !checkType( fileDesc->cf, "java/io/FileDescriptor" ) )
	{
		env->retcode = newCompleteException( env, InternalError, "Bad FileDescriptor object" );
		return;
	}

	fd = fileDesc->data[0];

	if( (fd < 1) || !fdList[fd-1] )
	{
		env->retcode = newCompleteException( env, IOException, "Bad file descriptor" );
		return;
	}

	switch( fd - 1 )
	{
	case 0: /* stdin */
	case 1: /* stdout */
	case 2: /* stderr */
		env->retcode = newCompleteException( env, IOException, "Close not permitted on in, out or err" );
		return;
	default:
		fclose( fdList[fd-1] );
		fdList[fd-1] = 0;
		break;
	}
}
